import importlib
import threading
import configparser
import pathlib
import pickle
import os
import MySQLdb

from jutil.installation import FileSetupController, SetupController, installation_database_singleton
from ScopusWp.config import TEMPLATE_PATH, Config, PATH


class ProjectPathInputController:

    def __init__(self):
        self._path = None

    @property
    def path(self):
        self.run()
        return self._path

    def run(self):
        if not self.check_config():
            self.prompt_path()

    def prompt_path(self):
        while True:
            inp = input('\nENTER PROJECT PATH: ')
            try:
                path = pathlib.Path(inp)
                if path.exists() and path.is_dir():
                    self._path = str(path)
                    break
                else:
                    print('INVALID PATH, TRY AGAIN')
                    continue
            except:
                print('INVALID INPUT, TRY AGAIN')
                continue

    def check_config(self):
        import ScopusWp.config as _config
        # In case the project path value is not empty, it is save to assume it has been set already and the path is
        # still valid
        if _config.PROJECT_PATH == '':
            print('NO PATH HAS BEEN SET')
            return False

        # Asking the user if he wants to keep that path or change it
        print('CURRENT PATH:\n{}'.format(_config.PROJECT_PATH))
        inp = input('\nDo you want to keep that path? Type "N" to change...')
        if inp.lower() == 'n' or inp.lower() == 'no':
            return False
        else:
            self._path = _config.PROJECT_PATH
            return True


def write_project_path(project_path):
    # Reading the config.py file of the package
    with open(os.path.join(PATH, 'config.py')) as file:
        lines = file.readlines()

    # Changing the line with the project path
    count = 0
    for i in range(len(lines)):
        line = lines[i]
        if 'PROJECT_PATH' in line:
            break
        count += 1

    lines[count] = 'PROJECT_PATH = "{}"'.format(project_path)
    # writing the new lines into the file
    with open(os.path.join(PATH, 'config.py'), mode='w+') as file:
        file.write('\n'.join(lines))


class PubControlSetup(SetupController):

    def __init__(self, project_path):
        # Getting the path of the whole thing
        self.path = project_path

        # The file setup for the config ini
        self.config_setup = FileSetupController(
            self.path,
            'config.ini',
            os.path.join(TEMPLATE_PATH, 'config.ini'),
            mysql_host='',
            mysql_username='',
            mysql_password='',
            mysql_database=''
        )

    def run(self):
        self.info()
        self.setup_files()
        self.setup_database()

    def setup_files(self):
        print('\n[PUBCONTROL FILE SETUP]')
        self.config_setup.run()

    def setup_database(self):
        print('\n[PUBCONTROL DATABASE SETUP]')
        try:
            class Db(installation_database_singleton(
                Config.get_instance()['MYSQL']['username'],
                Config.get_instance()['MYSQL']['password'],
                Config.get_instance()['MYSQL']['host']
            )):
                pass

            from ScopusWp.database import MySQLDatabase, BASE
            import ScopusWp.register

            # Creating the database
            Db.create_database('__pubcontrol')
            print('"__pubcontrol" DATABASE CREATED')
            # Creating all the tables according to the models
            session = MySQLDatabase.get_session()
            MySQLDatabase.create_database(BASE)
            print('PUBCONTROL DATABASE TABLES CREATED')

        except Exception as e:
            print('FAILED TO INSTALL SCOPUS DATA BASE DUE TO EXCEPTION: "{}"'.format(str(e).strip('\n')))

    @staticmethod
    def info():
        import ScopusWp.config as _config
        import os

        print('\nCURRENT VERSION:\n{}'.format(_config.VERSION))
        print('\nRUNNING IN FOLDER:\n{}'.format(_config.PATH))
        print('\nPACKAGE CONTENT:\n{}'.format(os.listdir(_config.PATH)))


def main():
    from ScopusWp.wordpress.install import WordpressSetup
    from ScopusWp.scopus.install import ScopusSetup

    project_path = ProjectPathInputController().path
    write_project_path(project_path)

    PubControlSetup(project_path).run()
    ScopusSetup(project_path).run()
    WordpressSetup(project_path).run()




from jutil.installation import SetupController, FileSetupController, installation_database_singleton

from ScopusWp.config import PATH, Config

import pathlib
import os


class ScopusSetup(SetupController):

    def __init__(self, project_path):
        self.path = pathlib.Path(project_path) / 'scopus'

        self.authors_setup = FileSetupController(
            str(self.path),
            'authors.ini',
            os.path.join(PATH, 'scopus', 'templates', 'authors.ini'),
            shorthand='SHORTHAND',
            ids=[],
            first_name='',
            last_name='',
            keywords=[],
            whitelist=[],
            blacklist=[]
        )

    def run(self):
        self.setup_files()
        self.setup_database()

    def create_folder(self):
        if not self.path.exists():
            self.path.mkdir()
            print('CREATED SCOPUS FOLDER IN PROJECT FOLDER')

    def setup_files(self):
        print('\n[SCOPUS FILE SETUP]')
        self.create_folder()

        self.authors_setup.run()

    def setup_database(self):
        print('\n[SCOPUS DATABASE SETUP]')
        try:
            class Db(installation_database_singleton(
                Config.get_instance()['MYSQL']['username'],
                Config.get_instance()['MYSQL']['password'],
                Config.get_instance()['MYSQL']['host']
            )):
                pass

            from ScopusWp.scopus.persistency import ScopusMySQLDatabase, BASE

            # Creating the database
            Db.create_database('__scopus')
            print('"__scopus" DATABASE CREATED')
            # Creating all the tables according to the models
            session = ScopusMySQLDatabase.get_session()
            ScopusMySQLDatabase.create_database(BASE)
            print('SCOPUS DATABASE TABLES CREATED')

            from ScopusWp.database import MySQLDatabase
            from ScopusWp.register import OriginType
            session = MySQLDatabase.get_session()
            session.add(OriginType(typeID=1, name='scopus'))
            session.commit()

        except Exception as e:
            print('FAILED TO INSTALL SCOPUS DATA BASE DUE TO EXCEPTION: "{}"'.format(str(e).strip('\n')))




from ScopusWp.scopus.data import ScopusPublication, ScopusAuthor, ScopusAuthorObservation

from ScopusWp.scopus.observe import ScopusObservationController

from ScopusWp.scopus.persistency import ScopusDatabaseController
from ScopusWp.scopus.persistency import ScopusDatabaseController

from ScopusWp.scopus.scopus import ScopusController

from ScopusWp.config import PATH
from ScopusWp.extend import DataSource

from ScopusWp.dat import GeneralPublication

import logging

# TODO: Implement massive logging


# TODO: Make a controller for temp Persistency
# Todo: Make the new methods use the get methods that also cache

# todo: Make new cache based on database cuz faster
# todo: Logging mal richtig angehen


class ScopusTopController(DataSource):

    def __init__(self, controller):
        DataSource.__init__(self, controller)

        self.observation_controller = ScopusObservationController()
        self.scopus_controller = ScopusController()
        self.database_controller = ScopusDatabaseController()

        self.logger = logging.getLogger('ScopusTop')

    #####################
    # TOP LEVEL METHODS #
    #####################

    def fetch(self):
        # Getting all the author observations
        author_observations = self.get_author_observations()
        print(author_observations)
        # fetching all the author profiles from the scopus website
        author_profiles = []
        for author_observation in author_observations:  # type: ScopusAuthorObservation

            for author_id in author_observation.ids:
                author_profile = self.get_author_profile(author_id, caching=True)
                author_profiles.append(author_profile)
        print(author_profiles)
        black_list = [42467747, 42737694, 42362046, 37612136, 36878195, 37136595, 36769475, 36769023, 36769945, 37194807, 41497708, 36706997, 37063140, 36591014, 702861, 37122867, 35960842, 35945051, 479546, 348142464, 503106, 347868461, 347868483, 35959943, 35940025, 2023475, 1648623, 2902277, 8495492, 41152093, 958889, 6922991, 34702161, 34702159, 34658985, 34659004, 34720088, 504571, 40017552, 1134957, 1566745, 10532766, 643029, 1070458, 33583412, 33583405, 33545152, 970760, 40676868, 1482169, 347811604, 42388612, 41386439, 37687439, 37151459, 36703153, 36703243, 35552798, 41356967, 36357725, 36030857, 35354081, 33721359, 32594078, 32238802, 32138032, 29012972, 345171814, 29233456, 33356885]
        for author_profile in author_profiles:

            for scopus_id in author_profile.publications:
                # Requesting the publication and in case there is an error during the request skipping the publication
                try:
                    if scopus_id not in black_list:
                        scopus_publication = self.get_publication(scopus_id, caching=True)
                except ConnectionError:
                    #black_list.append(scopus_id)
                    print(black_list)
                    continue
                # Deciding whether to actually return it
                if self.observation_controller.supports_publication(scopus_publication):

                    # If the publication is actually in the set of observed publications, getting all the citations
                    # as publications first and returning them
                    for citation_scopus_id in scopus_publication.citations:
                        try:
                            if citation_scopus_id not in black_list:
                                citation_publication = self.get_publication(citation_scopus_id, caching=True)
                                # Removing the citation, in an attempt not to overload the table, that saves the pending
                                # citations to be added
                                citation_publication.citations = []
                                publication_dict = self.general_publication(citation_publication)
                                publication_dict['origin']['text'] = 'citation'
                                yield publication_dict
                                continue
                        except ConnectionError:
                            #black_list.append(scopus_id)
                            print(black_list)
                            continue

                    yield self.general_publication(scopus_publication)

    def general_publication(self, scopus_publication):
        publication_dict = {
            'title': scopus_publication.title,
            'description': scopus_publication.description,
            'published': scopus_publication.date,
            'doi': scopus_publication.doi,
            'journal': scopus_publication.journal,
            'volume': scopus_publication.volume,
            'origin': {
                'name': self.name,
                'id': scopus_publication.id,
                'text': ''
            },
            'authors': list(map(
                lambda scopus_author: {
                    'first_name': scopus_author.first_name,
                    'last_name': scopus_author.last_name
                },
                scopus_publication.authors
            )),
            'citations': list(map(
                lambda scopus_id: {'id': scopus_id, 'name': self.name},
                scopus_publication.citations
            )),
            'links': [],
            'categories': self.observation_controller.get_publication_keywords(scopus_publication),
            'tags': scopus_publication.keywords
        }

        return publication_dict

    @property
    def name(self):
        return 'scopus'

    def explore_author_affiliations(self, author_dict):
        """
        Gets the list of all affiliation ids occurring in the history of all the publications to a author id of an
        author.

        Takes a dict, that assigns a list of possible author ids to a string tuple with last name and first name of
        an author
        Returns a dict with the string name tuple as keys and the values being dicts, that assigns list of affiliation
        ids to the author ids given in the input dict
        Example:
        IN: {('john', 'doe'): [1987623, 1294401]}
        OUT: {('john', 'doe'): {1987623: [8383992, 12387293],
                                1294401: [2312123]}
        :param author_dict: {('first name', 'last name') -> [author ids]}}
        :return: {('first name', 'last name') -> {author id -> [affiliation ids]}}
        """
        author_affiliation_dict = {}

        self.logger.info('loading temp storage for author aff.')

        # The naming function for the storage files
        def name_function(obj): return ''.join(list(obj.keys())[0])
        # Creating the temp storage list to save data persistently in case of crash
        temp_list = list('au_aff', PATH + '/temp', name_function)
        temp_list.load()

        # Loading the values, that were already saved in the temp list
        for temp_dict in temp_list:
            author_affiliation_dict.update(temp_dict)

        self.logger.info('requesting publications for affiliations')
        for name_tuple, author_id_list in author_dict.items():
            # Only really processing and requesting for a user, if that user is not already in the dict
            if name_tuple not in author_affiliation_dict.keys():
                # Getting the affiliation id list for each of the author ids and saving the list as the value to the key
                # being the author id
                affiliation_dict = {}
                for author_id in author_id_list:
                    affiliation_id_list = self.get_affiliations_author(author_id)
                    affiliation_dict[author_id] = affiliation_id_list
                # Adding the affiliation dict as the value to the name tuple key to the main dict
                temp_dict = {name_tuple: affiliation_dict}
                author_affiliation_dict.update(temp_dict)
                # Saving the temp dict, which represents the main entry for a single author
                temp_list.append(temp_dict)

        self.logger.info('finished exploring affiliations')
        return author_affiliation_dict

    def get_affiliations_author(self, author_id):
        """
        Requests the author profile for the author from the scopus system and then all the publications of that author.
        Goes through all publications and returns a list of all the different affiliation ids, which the author had
        in the history of all his publications.

        Uses caching on default.
        :param author_id: The id of the author for which to get the affiliations
        :return:
        """
        # Getting the author profile for the author id
        author_profile = self.get_author_profile(author_id)
        # Getting all the publications for the author
        publication_list = self.get_multiple_publications(author_profile.publications)

        affiliation_list = []
        for publication in publication_list:  # type: ScopusPublication
            # Getting the author entry from the publication object, that belongs to the author specified by the passed
            # author id
            if len(publication.authors) > 0:
                author = publication.authors[0]
                for author in publication.authors:  # type: ScopusAuthor
                    if int(author_profile) == int(author):
                        break

                difference = list(set(author.affiliations) - set(affiliation_list))
                affiliation_list += difference

        return affiliation_list

    def get_author_observations(self):
        """
        This method will return all the ScopusAuthorObservation objects. Those objects contain the data specified in
        the config file for which authors to be represented in the system and the website.

        :return: [ScopusAuthorObservation]
        """
        return self.observation_controller.all_observations()

    def get_multiple_publications(self, scopus_id_list, caching=True):
        """
        Gets multiple publications based on the multiple scopus ids given in the list.

        :param scopus_id_list: The list of scopus ids for which to get the publications
        :param caching: The boolean flag of whether to use caching or not
        :return: [ScopusPublication]
        """
        publication_list = []
        for scopus_id in scopus_id_list:
            publication = self.get_publication(scopus_id, caching=caching)
            publication_list.append(publication)
        return publication_list

    def get_publications_observed(self, caching=True):
        """
        Gets ALL the publications from the observed authors.

        :param caching: The boolean flag of whether to use caching or not
        :return: [ScopusPublication]
        """
        # Getting the author ids of all the observed authors for the scopus database
        author_id_list = self.observation_controller.all_observed_ids()
        # Getting the author profiles for all the authors
        author_profile_list = []
        for author_id in author_id_list:
            author_profile = self.get_author_profile(author_id, caching=caching)
            author_profile_list.append(author_profile)
        # Getting the publications for all the author profiles
        scopus_id_list = []
        for author_profile in author_profile_list:
            _scopus_id_list = author_profile.publications
            _difference = list(set(_scopus_id_list) - set(scopus_id_list))
            scopus_id_list += _difference
        if '' in scopus_id_list:
            scopus_id_list.remove('')

        publication_list = self.get_multiple_publications(scopus_id_list)

        return publication_list

    def get_publication_ids_observed(self, caching=True):
        """
        Gets the publication ids of all the publications of all the observed authors

        Gets all the author ids of the observed authors, gets all the author profiles for those ids and then gets from
        the profiles the lists of publications and adds those to the overall list, that have not already been added by
        a co author. returns the over all list.
        :param caching: The boolean flag of whether or not to use caching for the author profiles
        :return: [int] the scopus ids
        """
        # Getting the list of observed authors from the observation controller
        observed_author_id_list = self.observation_controller.all_observed_ids()

        publication_id_list = []
        for author_id in observed_author_id_list:
            # Attempting to get the author profiles for the publication id lists from the cache and if the cache
            # does not contain them, requesting them from scopus
            author_profile = self.get_author_profile(author_id, caching=caching)
            publication_difference_list = list(set(author_profile.publications) - set(publication_id_list))
            publication_id_list += publication_difference_list
            self.logger.info(
                'Fetched the author profile and author publication data for author {}-{}-{}'.format(
                    author_profile.id,
                    author_profile.first_name,
                    author_profile.last_name
                )
            )

        return publication_id_list

    def request_publication_ids_observed(self):
        # Getting the author ids of all the observed authors for the scopus database
        author_id_list = self.observation_controller.all_observed_ids()
        # Getting the author profiles for all the authors
        author_profile_list = self.scopus_controller.get_multiple_author_profiles(author_id_list)
        # Assembling the list of publications for each author profile
        scopus_id_list = []
        for author_profile in author_profile_list:
            difference = list(set(author_profile.publications) - set(scopus_id_list))
            scopus_id_list += difference
        return scopus_id_list

    def reload_cache_observed(self):

        # Loading the cache anew
        self.load_cache_observed()

    def load_cache_observed(self, load_citations=True):
        """
        First loads the observed authors profiles into the cache (requests only those not already in the cache).
        Based on those author profiles loads the publications of all the observed authors into the cache and if the
        load_citations flag is set (default), all the publications, that have cited those publications will also be
        loaded into the cache.

        :param load_citations: boolean flag of whether or not to load the citation publications into the cache as well
        :return: void
        """
        # Loading the author profiles into the cache
        self._load_cache_observed_authors()

        # Loading the publications of the observed authors into the cache
        self._load_cache_observed_publications(load_citations=load_citations)

    def load_publications_cache(self, scopus_id_list, auto_save_interval=10, reload=False):
        """
        Loads the publication info about the publications described by the scopus ids in the list from the scopus
        website and saves them in the cache.

        Although only specifically requests those publications from the web, that are not already in the cache, for
        network performance reasons. All the publications can be requested by setting reload to true.
        Also the cache auto saves after the specified amount for the auto save interval in case of connection error.

        :param scopus_id_list: The list of scopus ids, for all the publications to be loaded into the cache
        :param auto_save_interval: The amount of publications to be fetched from the scopus website, before the
            cache saves the progress. default on 20
        :param reload: The boolean value of whether or not to get all the specified publications from the scopus
            website or leave out those, already in the cache
        :return: void
        """
        # Getting the list of publication ids for those publications, that are still in the cache
        cache_scopus_id_list = self.database_controller.select_all_publication_ids()

        # If the reload flag is True, using the whole scopus id list as the list to be requested from the scopus site
        # else, only using those ids, that are not in the cache already
        if reload:
            difference_scopus_id_list = scopus_id_list
        else:
            difference_scopus_id_list = list(set(scopus_id_list) - set(cache_scopus_id_list))

        auto_save_count = 0
        for scopus_id in difference_scopus_id_list:
            # Auto saving if the count has reached the specified interval and then resetting the counter
            if auto_save_count == auto_save_interval:
                self.database_controller.save()
                auto_save_count = 0

            # Getting the publication from the scopus website
            publication = self.scopus_controller.get_publication(scopus_id)
            self.database_controller.insert_publication(publication)

            auto_save_count += 1

        self.database_controller.save()

    def load_authors_cache(self, author_id_list, auto_save_interval=10, reload=False):
        """
        Loads the author profiles for the authors given by the author id list into the cache, by requesting them
        from the scopus website.

        On default, only those author profiles, that cannot already be found in the cache will be explicitly
        requested from the scopus website. When reload is True, all the author profiles will be requested and the
        possibly existing cache will be overwritten.
        Also the cache auto saves after the specified amount for the auto save interval, in case of connection
        error or exception.

        :param author_id_list: The list of author ids for the author profiles to be loaded into the cache
        :param auto_save_interval: The int amount of author profiles to be requested before the cache saves the
            progress into persistency.
        :param reload: The boolean value of whether or not the
        :return: void
        """
        cache_author_id_list = self.database_controller.select_all_author_ids()

        # Subtracting the the observed from the cached ids and adding only the left over to the cache
        if reload:
            difference_author_id_list = author_id_list
        else:
            difference_author_id_list = list(set(author_id_list) - set(cache_author_id_list))

        auto_save_counter = 0
        for author_id in difference_author_id_list:
            # Saving during the process, so progress is not lost after connection error
            if auto_save_counter == auto_save_interval:
                self.database_controller.save()
                auto_save_counter = 0

            # Requesting the author profile from the scopus website
            author_profile = self.scopus_controller.get_author_profile(author_id)
            self.database_controller.insert_author_profile(author_profile)
            auto_save_counter += 1
        self.database_controller.save()

    def _load_cache_observed_authors(self):
        """
        Loads the author profiles for all the observed authors into the cache, but only specifically requests those,
        which are not already in the cache.

        :return: void
        """
        # Getting the list of author ids for the observed authors from the observation controller
        observed_author_id_list = self.observation_controller.all_observed_ids()

        # Loading all those author profiles of the observed authors into the cache
        self.load_authors_cache(observed_author_id_list)

    def _load_cache_observed_publications(self, load_citations=True):
        """
        Loads all the publications of the observed authors into the cache, but only those, that are not already in the
        cache. Also loads all the publication, that have cited the observed publications into the cache if enabled.

        :param load_citations: boolean flag of whether or not to also load the citation publications into the cache
        :return: void
        """
        # Getting the list of publication ids for all the observed authors from the cache
        observed_publication_id_list = self.get_publication_ids_observed()

        # Loading all those publications into the cache after having them requested from the scopus website
        self.load_publications_cache(observed_publication_id_list)

        if load_citations:
            # Getting the citations scopus id list from every one of those publications
            citation_scopus_id_list = []
            for scopus_id in observed_publication_id_list:
                publication = self.database_controller.select_publication(scopus_id)

                # Protecting against the unlikely case of corrupted data with double checking for None
                if publication is not None:
                    difference = list(set(publication.citations) - set(citation_scopus_id_list))
                    citation_scopus_id_list += difference

            # Loading all those publications into the cache as well
            self.load_publications_cache(citation_scopus_id_list)

    def get_citation_publications(self, publication):
        publication_list = self.scopus_controller.get_citation_publications(publication)
        # Caching all the citation publications
        for publication in publication_list:
            self.database_controller.insert_publication(publication)
        return publication_list

    def populate_database(self, author_id_list):
        """
        Given a list with author ids, this method will load the author ids as author profiles and then their
        publications into the scopus database

        :param author_id_list: A list of int author ids
        :return: void
        """
        author_profile_list = list(map(lambda x: self.get_author_profile(x), author_id_list))
        # Getting the publications of those authors
        publication_list = []
        for author_profile in author_profile_list:
            publication_list += list(map(lambda x: self.get_publication(x), author_profile.publications))

    def get_author_profile(self, author_id, caching=True):
        """
        Returns the author profile for the given author id.
        Uses caching if flag is set. so tries to get from cache first and if not in cache gets from the scopus website
        and then writes back into the cache.

        :param author_id: The int author id for the author
        :param caching: boolean flag if use caching or not
        :return: ScopusAuthorProfile
        """
        # Checking if the author profile is already in the cache
        is_cached = self.database_controller.contains_author_profile(author_id)
        if is_cached and caching:
            author_profile = self.database_controller.select_author_profile(author_id)
        else:
            # Actually requesting from scopus website and then writing back into the cache
            author_profile = self.scopus_controller.get_author_profile(author_id)
            if not is_cached:
                self.database_controller.insert_author_profile(author_profile)
        return author_profile

    def get_publication(self, scopus_id, caching=True):
        """
        Returns the publication object for the given scopus id.
        Uses caching if flag is set. So tries to get the publication from the cache first, if not in the cache, gets
        the publication from the scopus website and then writes back into the cache.

        :param scopus_id: The in scopus id identifying the publication
        :param caching: boolean flag if caching enbled or not
        :return: ScopusPublication
        """
        # Checking if the publication is in the cache and returning the cached value if possible
        is_cached = self.database_controller.contains_publication(scopus_id)
        if is_cached and caching:
            publication = self.database_controller.select_publication(scopus_id)
        else:
            # If the publication is not cached, getting it from the scopus website
            publication = self.scopus_controller.get_publication(scopus_id)
            if not is_cached:
                # And then writing it into the cache for the next time
                self.database_controller.insert_publication(publication)
        return publication

    def save(self):
        self.database_controller.save()

    ######################
    # THE SCOPUS METHODS #
    ######################

    def request_publication(self, scopus_id):
        return self.scopus_controller.get_publication(scopus_id)

    def request_multiple_publications(self, scopus_id_list):
        return self.scopus_controller.get_multiple_publications(scopus_id_list)

    ###########################
    # THE OBSERVATION METHODS #
    ###########################

    def filter_by_observation(self, publication_list):
        return self.observation_controller.filter(publication_list)

    def publication_observation_keywords(self, publication):
        return self.observation_controller.get_publication_keywords(publication)

    ######################
    # THE DATAB  METHODS #
    ######################

    def select_publication_database(self, scopus_id):
        return self.database_controller.select_publication(scopus_id)

    def select_multiple_publications_database(self, scopus_id_list):
        return self.database_controller.select_multiple_publications(scopus_id_list)

    def select_all_publications_database(self):
        return self.database_controller.select_all_publications()

    def insert_publication_database(self, publication):
        self.database_controller.insert_publication(publication)

    def insert_multiple_publication_database(self, publication_list):
        self.database_controller.insert_multiple_publications(publication_list)




from jutil.installation import SetupController, FileSetupController, installation_database_singleton

from ScopusWp.config import PATH, Config

import pathlib
import os


class WordpressSetup(SetupController):

    def __init__(self, project_path):
        SetupController.__init__(self)

        self.path = pathlib.Path(project_path) / 'wordpress'

        self.config_setup = FileSetupController(
            str(self.path),
            'config.ini',
            os.path.join(PATH, 'wordpress', 'templates', 'config.ini'),
            url='',
            username='',
            password=''
        )

    def run(self):
        self.setup_files()
        self.setup_database()

    def setup_database(self):
        print('\n[WORDPRESS DATABASE SETUP]')
        try:
            class Db(installation_database_singleton(
                Config.get_instance()['MYSQL']['username'],
                Config.get_instance()['MYSQL']['password'],
                Config.get_instance()['MYSQL']['host']
            )):
                pass

            from ScopusWp.wordpress.reference import WordpressMySQLDatabase, BASE

            # Creating the database
            Db.create_database('__wordpress')
            print('"__wordpress" DATABASE CREATED')
            # Creating all the tables according to the models
            session = WordpressMySQLDatabase.get_session()
            WordpressMySQLDatabase.create_database(BASE)
            print('WORDPRESS DATABASE TABLES CREATED')

        except Exception as e:
            print('FAILED TO INSTALL SCOPUS DATA BASE DUE TO EXCEPTION: "{}"'.format(str(e).strip('\n')))

    def setup_files(self):
        print('\n[WORDPRESS FILE SETUP]')
        self.create_folder()

        self.config_setup.run()

    def create_folder(self):
        if not self.path.exists():
            self.path.mkdir()
            print('CREATED WORDPRESS FOLDER IN PROJECT FOLDER')



from ScopusWp.config import PROJECT_PATH

from jutil.configuration import config_singleton

import os


PATH = os.path.dirname(os.path.realpath(__file__))

class WordpressConfig(config_singleton(
    PROJECT_PATH + '/wordpress',
)):
    pass


from ScopusWp.extend import DataSink

from ScopusWp.wordpress.wordpress import WordpressPublicationController
from ScopusWp.wordpress.reference import PublicationReference


class Wordpress(DataSink):

    def __init__(self, controller):
        DataSink.__init__(self, controller)

        self.reference = PublicationReference()
        self.wordpress_controller = WordpressPublicationController()

    def run(self):
        self.reference.wipe()
        # Selecting all the publications from scopus
        general_publications = self.register.select_origin('scopus', '')

        # Posting all the publications, that are not yet in the system to the web page
        for general_publication in general_publications:
            if not self.reference.contains_publication(general_publication.id):
                post_reference_dict = self.wordpress_controller.post(general_publication)
                self.reference.insert(post_reference_dict)

    @property
    def name(self):
        return 'wordpress'



from ScopusWp.database import MySQLDatabaseAccess

from ScopusWp.scopus.data import ScopusPublication, ScopusAuthorProfile, ScopusAuthor
from ScopusWp.scopus.data import from_dict, to_dict

from ScopusWp.config import PATH, Config

import json
import pathlib

import pickle
import os
import datetime

from sqlalchemy import Column, ForeignKey, Integer, String, Text, DATETIME, BigInteger, Table, and_
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, Session

from ScopusWp.database import MySQLDatabase, get_or_create
from ScopusWp.util import IDManager

from jutil.database import *

BASE = declarative_base()


class ScopusMySQLDatabase(mysql_database_singleton(
    Config.get_instance()['MYSQL']['username'],
    Config.get_instance()['MYSQL']['password'],
    Config.get_instance()['MYSQL']['host'],
    '__scopus'
)):
    pass


publication_snapshot_junction = Table(
    'publication_snapshot_junction',
    BASE.metadata,
    Column('authorID', BigInteger, ForeignKey('author_snapshot.snapshotID'), primary_key=True),
    Column('scopusID', BigInteger, ForeignKey('publication.scopusID'), primary_key=True)
)


class Author(BASE):

    __tablename__ = 'author'

    id = Column(BigInteger, primary_key=True)
    first_name = Column(String(500))
    last_name = Column(String(500))
    h_index = Column(Integer)
    citation_count = Column(Integer)
    publications = Column(Text)
    document_count = Column(Integer)
    affiliation_country = Column(String(100))
    affiliation_city = Column(String(100))
    affiliation_name = Column(String(300))


class AuthorSnapshot(BASE):

    __tablename__ = 'author_snapshot'

    snapshotID = Column(BigInteger, primary_key=True)
    author_id = Column(BigInteger)
    first_name = Column(String(500))
    last_name = Column(String(500))
    affiliations = Column(Text)
    type = Column(String(100))

    publications = relationship(
        'Publication',
        secondary=publication_snapshot_junction,
        back_populates='authors'
    )


class Publication(BASE):

    __tablename__ = 'publication'

    scopusID = Column(BigInteger, primary_key=True)
    eid = Column(String(100))
    doi = Column(String(100))
    creator_id = Column(BigInteger, ForeignKey('author_snapshot.snapshotID'))
    title = Column(Text)
    description = Column(Text)
    journal = Column(String(500))
    volume = Column(String(100))
    date = Column(DATETIME)
    keywords = Column(Text)
    citations = Column(Text)

    creator = relationship(
        'AuthorSnapshot',
        uselist=False,
    )

    authors = relationship(
        'AuthorSnapshot',
        secondary=publication_snapshot_junction,
        back_populates='publications'
    )


class ScopusDatabaseController:

    def __init__(self):

        self.session = ScopusMySQLDatabase.get_session()  # type: Session

        self.author_id_manager = IDManager('author_snapshot')

    def insert_author_profile(self, author_profile):
        author = get_or_create(
            self.session,
            Author,
            id=author_profile.id,
            first_name=author_profile.first_name,
            last_name=author_profile.last_name,
            h_index=author_profile.h_index,
            citation_count=author_profile.citation_count,
            publications=json.dumps(author_profile.publications),
            document_count=author_profile.document_count,
            affiliation_country="",
            affiliation_city="",
            affiliation_name=""
        )
        self.session.commit()

    def contains_author_profile(self, author):
        return len(list(self.session.query(Author).filter(Author.id == int(author)))) != 0

    def select_author_profile(self, author_id):
        results = self.session.query(Author).filter(Author.id == int(author_id))
        author = results[0]

        author_profile = ScopusAuthorProfile(
            author.id,
            author.first_name,
            author.last_name,
            author.h_index,
            author.citation_count,
            author.document_count,
            json.loads(author.publications)
        )

        return author_profile

    def select_all_author_profiles(self):
        return self.session.query(Author).all()

    def insert_publication(self, publication):

        pub = Publication(
            scopusID=publication.id,
            eid=publication.eid,
            doi=publication.doi,
            title=publication.title,
            description=publication.description,
            journal=publication.journal,
            volume=publication.volume,
            date=datetime.datetime.strptime(publication.date, '%Y-%M-%d'),
            keywords=json.dumps(publication.keywords),
            citations=json.dumps(publication.citations)
        )

        creator_snapshot = get_or_create(
            self.session,
            AuthorSnapshot,
            author_id=publication.creator.id,
            first_name=publication.creator.first_name,
            last_name=publication.creator.last_name,
            affiliations=json.dumps(sorted(publication.creator.affiliations)),
            type='ScopusAuthor'
        )
        pub.creator = creator_snapshot

        author_snapshots, author_snapshots_remaining = sqlalchemy_many_list_tuple(
            self.session,
            AuthorSnapshot,
            publication.authors,
            lambda author: and_(AuthorSnapshot.author_id == int(author),
                                AuthorSnapshot.affiliations == json.dumps(sorted(author.affiliations))),
            lambda author: AuthorSnapshot(
                author_id=author.id,
                first_name=author.first_name,
                last_name=author.last_name,
                affiliations=json.dumps(sorted(author.affiliations)),
                type='ScopusAuthor'
            )
        )

        pub.authors = author_snapshots
        self.session.add(pub)

        self.session.commit()

    def _insert_author_snapshot(self, author):

        author_snapshot_id = self.author_id_manager.new()
        author_snapshot = get_or_create(
            self.session,
            AuthorSnapshot,
            id=author_snapshot_id,
            author_id=author.id,
            first_name=author.first_name,
            last_name=author.last_name,
            affiliations=json.dumps(sorted(author.affiliations)),
            type='ScopusAuthor'
        )

        return author_snapshot_id

    def _select_author_snapshot(self, snapshot_id):
        results = self.session.query(AuthorSnapshot).filter(AuthorSnapshot.snapshotID == snapshot_id)
        # Creating the actual ScopusAuthor object from that
        snapshot = results[0]  # type: AuthorSnapshot
        scopus_author = ScopusAuthor(
            snapshot.first_name,
            snapshot.last_name,
            snapshot.snapshotID,
            json.loads(snapshot.affiliations)
        )
        return scopus_author

    def insert_multiple_publications(self, publication_list):
        for publication in publication_list:
            self.insert_publication(publication)

    def contains_publication(self, publication):
        return len(list(self.session.query(Publication).filter(Publication.scopusID == int(publication))))

    def _scopus_author_from_snapshot(self, author_snapshot):
        return ScopusAuthor(
            author_snapshot.first_name,
            author_snapshot.last_name,
            author_snapshot.snapshotID,
            json.loads(author_snapshot.affiliations)
        )

    def select_publication(self, scopus_id):
        results = self.session.query(Publication).filter(Publication.scopusID == int(scopus_id))
        publication = results[0]
        # Loading the creator and the authors list from the snapshots database

        creator = self._scopus_author_from_snapshot(publication.creator)
        author_list = list(map(
            lambda x: self._scopus_author_from_snapshot(x),
            publication.authors
        ))

        # Creating an actual ScopusPublication object from that
        scopus_publication = ScopusPublication(
            publication.scopusID,
            publication.eid,
            publication.doi,
            publication.title,
            publication.description,
            publication.date,
            creator,
            author_list,
            json.loads(publication.citations),
            json.loads(publication.keywords),
            publication.journal,
            publication.volume
        )
        return scopus_publication

    def wipe(self):
        #self.session.query(Author).delete()
        self.session.execute('TRUNCATE TABLE publication_snapshot_junction')
        self.session.query(Publication).delete()
        self.session.query(AuthorSnapshot).delete()
        self.session.commit()

    def select_multiple_publications(self, scopus_id_list):
        publication_list = []
        for scopus_id in scopus_id_list:
            publication = self.select_publication(scopus_id)
            publication_list.append(publication)
        return publication_list

    def select_all_publications(self):
        return self.session.query(Publication).all()

    def select_all_publication_ids(self):
        return self.session.query(Publication.scopusID).all()

    def select_all_author_ids(self):
        return self.session.query(Author.id).all()

    def save(self):
        self.session.commit()
        self.author_id_manager.save()

if __name__ == '__main__':
    session = ScopusMySQLDatabase.get_session()
    BASE.metadata.drop_all(ScopusMySQLDatabase._engine)
    BASE.metadata.create_all(ScopusMySQLDatabase._engine)



from ScopusWp.controller import TopController

import optparse

# TODO: Make a fail proof for the citation comment wordpress not well formed
# TODO: Make the new citations functionality with update date
# TODO: Make update date into the main reference
# TODO: Make error safe for http request pool exceptions wrapper method for requests



from ScopusWp.config import PATH, PROJECT_PATH
from ScopusWp.config import Config

from ScopusWp.database import MySQLDatabaseAccess

from ScopusWp.dat import GeneralPublication
from ScopusWp.util import IDManager

import datetime

import json
import pathlib

from sqlalchemy import Column, ForeignKey, Integer, String, Text, DATETIME, BigInteger
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, Session

from ScopusWp.database import MySQLDatabase, get_or_create
from ScopusWp.util import IDManager

Base = declarative_base()


DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'


class PublicationReference(Base):

    __tablename__ = 'publication_reference'

    id = Column(BigInteger, primary_key=True)
    wordpress_id = Column(BigInteger)
    scopus_id = Column(BigInteger)
    update_datetime = Column(DATETIME)
    post_datetime = Column(DATETIME)


class CommentReference(Base):

    __tablename__ = 'comment_reference'

    id = Column(BigInteger, primary_key=True)
    wordpress_post_id = Column(BigInteger)
    wordpress_comment_id = Column(BigInteger)
    external_id = Column(BigInteger)
    type = Column(Integer)
    posted_datetime = Column(DATETIME)


class ReferenceType(Base):

    __tablename__ = 'reference_type'

    id = Column(Integer, primary_key=True)
    name = Column(String(200))


class ReferenceController:

    def __init__(self):

        self.publication_reference_id_manager = IDManager('reference')
        self.session = MySQLDatabase.get_session()  # type: Session

    def select_reference(self, internal_id):
        return self.session.query(PublicationReference).filter(PublicationReference.id == internal_id)[0]

    def select_all_references(self):
        """
        Returns a list of tuples, where each tuple represents one entry in the reference database, with the first
        item being the internal id for the publication, the second being the wordpress id and the third being the
        scopus id for the publication.

        :return: A list of tuples with three items each
        """
        return list(self.session.query(PublicationReference).all())

    def insert_reference(self, internal_id, wordpress_id, scopus_id):
        ref = get_or_create(
            self.session,
            PublicationReference,
            id=internal_id,
            wordpress_id=wordpress_id,
            scopus_id=scopus_id,
            update_datetime=datetime.datetime.strptime('0001-01-01', '%Y-%M-%d'),
            post_datetime=datetime.datetime.now()
        )
        self.session.commit()

    def insert_comment_reference(self, internal_id, wordpress_post_id, wordpress_comment_id, scopus_id):
        """
        Inserts a new comment reference into the comment reference database

        :param internal_id: The internally created id for the post
        :param wordpress_post_id: The id of the wordpress post, the comment is added to
        :param wordpress_comment_id: The comment id specifically
        :param scopus_id: The scopus id of the publication on which the comment is based on
        :return: void
        """
        ref = get_or_create(
            self.session,
            CommentReference,
            id=internal_id,
            wordpress_post_id=wordpress_post_id,
            wordpress_comment_id=wordpress_comment_id,
            external_id=scopus_id,
            type=1,
            posted_datetime=datetime.datetime.now()
        )
        self.session.commit()

    def select_comment_reference_list_py_post(self, wordpress_post_id):
        """
        Gets a list of all the comment references for the post specified by the wordpress id

        :param wordpress_post_id: The int id of the wordpress post
        :return: [(internal id, post id, comment id, scopus id)]
        """
        return list(self.session.query(CommentReference).filter(CommentReference.wordpress_post_id==wordpress_post_id))

    def insert_publication(self, publication, wordpress_id, scopus_id):
        assert isinstance(publication, GeneralPublication)

        self.insert_reference(publication.id, wordpress_id, scopus_id)

    # TODO: This is not separation of concerns, what is this doing here
    def publication_from_scopus(self, scopus_publication):
        # Getting an id from the id manager
        publication_id = self.publication_reference_id_manager.new()

        # Creating the publication object
        publication = GeneralPublication.from_scopus_publication(scopus_publication, publication_id)
        return publication

    def select_post_reference_by_scopus(self, scopus_id):
        """
        Select the reference by the scopus id.

        :param scopus_id: The int scopus id
        :return: (internal id, wordpress id, scopus id)
        """
        return list(
            self.session.query(PublicationReference).filter(
                PublicationReference.scopus_id == scopus_id
            )
        )

    def select_post_reference_by_wordpress(self, wordpress_post_id):
        """
        Select the reference by the wordpress id of the post.

        :param wordpress_post_id: The int wordpress id
        :return: (internal id, wordpress id, scopus id)
        """
        return list(
            self.session.query(PublicationReference).filter(
                PublicationReference.wordpress_id == wordpress_post_id
            )
        )

    def wipe(self):
        self.session.query(PublicationReference).delete()
        self.session.query(CommentReference).delete()
        self.session.commit()

    def close(self):
        self.publication_reference_id_manager.save()


from ScopusWp.config import PATH, PROJECT_PATH

from ScopusWp.scopus.data import ScopusAuthorObservation

import os
import json
import configparser

#############
#  CLASSES  #
#############


class AuthorObservationInterface:

    def get_observation(self, author):
        raise NotImplementedError()

    def contains_author(self, author):
        raise NotImplementedError()

    def values(self):
        raise NotImplementedError()

    def keys(self):
        raise NotImplementedError()

    def __getitem__(self, item):
        raise NotImplementedError()

    def __contains__(self, item):
        raise NotImplementedError()

###############
# Controllers #
###############


class ScopusObservationController:

    def __init__(self):

        self.author_observation_model = AuthorObservationModel()

    def get_author_keywords(self, author):
        author_observation = self.author_observation_model[int(author)]  # type: ScopusAuthorObservation
        return author_observation.keywords

    def all_observations(self):
        return self.author_observation_model.values()

    def all_observed_ids(self):
        return self.author_observation_model.keys()

    def get_publication_keywords(self, publication):
        keywords = []
        author_observation_id_set = set(map(lambda x: int(x), self.author_observation_model.keys()))
        publication_author_id_set = set(map(lambda x: int(x), publication.authors))

        publication_observed_authors = list(publication_author_id_set.intersection(author_observation_id_set))
        # Going through all the authors of the publication and adding all the keywords of each observed author found
        for author in publication_observed_authors:
            author_keywords = self.get_author_keywords(author)
            difference = list(set(author_keywords) - set(keywords))
            keywords += difference
        return keywords

    def supports_publication(self, publication):
        """
        Returns the boolean value if the passed ScopusPublication is supported by the observation system, which means
        that at least one of its authors at the time of writing the publication had a affiliation, that is 7
        specifically whitelisted in the authors.ini configuration file for the observation.

        :param publication: ScopusPublication object to be checked
        :return: bool
        """
        for author in publication.authors:
            author_id = int(author)
            if author_id in self.author_observation_model.keys():
                author_observation = self.author_observation_model[author_id]  # type: ScopusAuthorObservation
                is_whitelist = author_observation.whitelist_contains_any(author.affiliations)
                # If the publication setting with author and affiliation is in at least one whitelist true can be
                # returned
                if is_whitelist:
                    return True

        # There is no instant False return for blacklist, because a potential whitelist entry of another author has
        # more priority, than a blacklist entry.
        # False is only returned in case True was not already returned due to a whitelisting
        return False

    def filter(self, publication_list):
        whitelist_publications = []
        blacklist_publications = []
        remaining_publications = []

        for publication in publication_list:
            _is_blacklist = False
            _is_whitelist = False
            # Checking the affiliations for all the authors
            for author in publication.authors:
                author_id = int(author.id)
                if author_id in self.author_observation_model.keys():
                    author_observation = self.author_observation_model[author_id] # type: ScopusAuthorObservation
                    _is_whitelist = author_observation.whitelist_contains_any(author.affiliations)
                    # If there was a blacklist found it would be overwritten with the next author, like this a TRUE
                    # will be preserved until the end of the loop
                    if not _is_blacklist:
                        _is_blacklist = author_observation.blacklist_contains_any(author.affiliations)

                    # In the case of whitelist, the publication will be added instantly, but for blacklist it is
                    # decided after all authors have been iterated
                    if _is_whitelist:
                        whitelist_publications.append(publication)
                        break

            if _is_blacklist:
                blacklist_publications.append(publication)
            else:
                remaining_publications.append(publication)

        return whitelist_publications, blacklist_publications, remaining_publications



############
#  Models  #
############


class AuthorObservationModel:

    def __init__(self):
        self.path_string = PROJECT_PATH + '/scopus/authors.ini'
        self.config = configparser.ConfigParser()
        self.config.read(self.path_string)
        print(self.path_string)

        self.dict, self.author_observations = self.load()

    def __contains__(self, item):
        self.contains_author(item)

    def __getitem__(self, item):
        return self.get_observation(item)

    def get_observation(self, author):
        return self.dict[int(author)]

    def contains_author(self, author):

        return int(author) in self.keys()

    def keys(self):
        return list(self.dict.keys())

    def values(self):
        return self.author_observations

    def load(self):
        content_dict = {}
        author_observation_list = []
        for section in self.config.keys():
            if section == 'DEFAULT':
                continue

            sub_dict = dict(self.config[section])

            # Getting the author observation
            author_observation = self._get_author_observation(sub_dict)
            author_observation_list.append(author_observation)
            # Adding a reference to the same author observation to all the author ids in the list
            for author_id in author_observation.ids:
                content_dict[int(author_id)] = author_observation

        return content_dict, author_observation_list

    @staticmethod
    def _get_author_observation(sub_dict):

        scopus_ids_json_string = sub_dict['ids']
        scopus_ids = json.loads(scopus_ids_json_string)

        keywords_json_string = sub_dict['keywords']
        keywords = json.loads(keywords_json_string)

        whitelist_json_string = sub_dict['scopus_whitelist']
        whitelist = json.loads(whitelist_json_string)

        blacklist_json_string = sub_dict['scopus_blacklist']
        blacklist = json.loads(blacklist_json_string)

        author_observation = ScopusAuthorObservation(
            scopus_ids,
            sub_dict['first_name'],
            sub_dict['last_name'],
            keywords,
            whitelist,
            blacklist
        )

        return author_observation


class ObservedAuthorsModel2:

    def __init__(self):
        self.path = os.path.join(PROJECT_PATH, 'authors2.ini')

        self.source = configparser.ConfigParser()
        self.source.read(self.path)

        self.author_observations = []
        self.author_ids = []

        self.author_observation_dict = {}
        # TODO: eventually dict struture for categories and ids to observ. objects

        for key in self.source.keys():
            if key == 'DEFAULT':
                continue
            # Each key represents one observed author
            value_dict = dict(self.source[key])

            author_observation = self._get_author_observation(value_dict)
            author_id_list = author_observation.ids
            self.author_ids += author_id_list
            self.author_observations.append(author_observation)
            for author_id in author_id_list:
                self.author_observation_dict[author_id] = author_observation

    def filter(self, publication_list):
        whitelist_publications = []
        blacklist_publications = []
        remaining_publications = []

        for publication in publication_list:
            _is_blacklist = False
            _is_whitelist = False
            # Checking the affiliations for all the authors
            for author in publication.authors:
                author_id = int(author.id)
                if author_id in self.author_ids:
                    author_observation = self.author_observation_dict[author_id]
                    _is_whitelist = author_observation.check_whitelist(author.affiliations)
                    # If there was a blacklist found it would be overwritten with the next author, like this a TRUE
                    # will be preserved until the end of the loop
                    if not _is_blacklist:
                        _is_blacklist = author_observation.check_blacklist(author.affiliations)

                    # In the case of whitelist, the publication will be added instantly, but for blacklist it is
                    # decided after all authors have been iterated
                    if _is_whitelist:
                        whitelist_publications.append(publication)
                        break

            if _is_blacklist:
                blacklist_publications.append(publication)
            else:
                remaining_publications.append(publication)

        return whitelist_publications, blacklist_publications, remaining_publications

    def get_observed_authors(self):
        return self.author_ids

    def _get_author_observation(self, value_dict):
        (
            first_name,
            last_name,
            author_id_list,
            keyword_list,
            scopus_whitelist,
            scopus_blacklist
        ) = self._extract_value_dict(value_dict)

        # Creating a new AuthorObservation object
        author_observation = AuthorObservation(
            author_id_list,
            first_name,
            last_name,
            keyword_list,
            scopus_whitelist,
            scopus_blacklist
        )

        return author_observation

    def _extract_value_dict(self, value_dict):
        first_name = value_dict['first_name']
        last_name = value_dict['last_name']

        author_ids_json = value_dict['ids']
        author_id_list = json.loads(author_ids_json)

        scopus_whitelist_json = value_dict['scopus_whitelist']
        scopus_whitelist = json.loads(scopus_whitelist_json)

        scopus_blacklist_json = value_dict['scopus_blacklist']
        scopus_blacklist = json.loads(scopus_blacklist_json)

        keyword_list_json = value_dict['keywords']
        keyword_list = json.loads(keyword_list_json)

        return first_name, last_name, author_id_list, keyword_list, scopus_whitelist, scopus_blacklist

    def __contains__(self, item):
        """
        The magic method for the 'xx in xx' Operator. If given an Author / AuthorProfile object or a author id string
        or int, checks and returns whether that author described by the author id is in the set of observed author ids

        :param item: Author/AuthorProfile/string/int anything that describes an author
        :return: The boolean value of whether or not that author is with the observed authors
        """
        return int(item) in self.author_ids

    def __getitem__(self, item):
        """
        Returns the AuthorObservation object for the given author

        :param item: Author/AuthorProfile/string/int anything that describes an author
        :return: The AuthorObservation object for the given author
        """
        # Works with the author id given directly as string/int and with the Author/AuthorProfile objects
        author_id = int(item)
        return self.author_observation_dict[author_id]  # type: AuthorObservation


import logging


class DataSource:

    def __init__(self, controller):
        self.controller = controller
        self.logger = logging.getLogger('source-{}'.format(self.name))

    def fetch(self):
        raise NotImplementedError()

    @property
    def name(self):
        raise NotImplementedError()


class DataSink:

    def __init__(self, controller):
        self.controller = controller
        self.logger = logging.getLogger('sink-{}'.format(self.name))

    @property
    def run(self):
        raise NotImplementedError()

    @property
    def register(self):
        return self.controller.register

    @property
    def config(self):
        return self.controller.config

    @property
    def name(self):
        raise NotImplementedError()



import json
from unidecode import unidecode

# TODO: Make the to_dict and from_dict accept lists

###############
#  FUNCTIONS  #
###############


def to_dict(obj):
    """
    Turns every object, that implements the DictConversionInterface into a dict.

    :param obj: The scopus data type, that converts into a dict
    :return: a dict object built from the data of the given object
    """
    if issubclass(obj.__class__, DictionaryConversionInterface):
        return obj.to_dict()
    elif isinstance(obj, list):
        dict_list = []
        for item in obj:
            _dict = to_dict(item)
            dict_list.append(_dict)
        return dict_list


def from_dict(scopus_dict):
    """
    Creates a object from the dict, that was created by the DictConversionInterface.

    :param scopus_dict: The dict to create an object from
    :return: An object of the class, that implements a DictConversionInterface
    """
    if isinstance(scopus_dict, dict):
        if 'type' in scopus_dict:
            code_string = (
                '{}.from_dict(scopus_dict)'
            ).format(scopus_dict['type'])
            scopus_object = eval(code_string)

            return scopus_object

    elif isinstance(scopus_dict, list):
        object_list = []
        for dictionary in scopus_dict:
            _object = from_dict(dictionary)
            object_list.append(_object)
        return object_list

###############
#   CLASSES   #
###############


class ScopusIdentifierInterface:
    """
    Interface, that promises, that when implemented the identifier of an object (id) can be acquired by calling an
    integer conversion on the object, or calling the 'get_id' method.
    """
    def get_id(self):
        raise NotImplementedError()

    def __int__(self):
        raise NotImplementedError()


class DictionaryConversionInterface:
    """
    Interface, that promises, that the implementing object can be converted into a dict and also that objects can be
    loaded/ created from such dicts.
    """
    def to_dict(self):
        """
        Supposed to convert all the data stored in the object into a dictionary.

        :return: A dict object
        """
        raise NotImplementedError()

    @staticmethod
    def from_dict(dictionary):
        """
        A static method supposed to be able to load objects of that class from a dictionary object

        :param dictionary: The dict object, containing the data about the construction of an object of that type
        :return: A class of the type, that implements this interface
        """
        raise NotImplementedError()


class ScopusPublication(ScopusIdentifierInterface, DictionaryConversionInterface):
    """
    Object, that represents a publication, that was retrieved from the Scopus database.
    """
    def __init__(self, scopus_id, eid, doi, title, description, date, creator, author_list, citation_list,
                 keyword_list, journal, volume):
        # Init the interfaces
        # ScopusIdentifierInterface.__int__(self)
        DictionaryConversionInterface.__init__(self)

        self.id = scopus_id
        self.eid = eid
        self.doi = doi
        self.title = unidecode(title).replace('"', "'")
        self.description = unidecode(description).replace('"', "'")
        self.date = date
        self.creator = creator
        self.authors = author_list
        self.citations = citation_list
        if '' in self.citations:
            self.citations.remove('')
        self.keywords = list(map(self._rip, keyword_list))
        self.journal = unidecode(journal).replace('"', "'")
        self.volume = volume

    @staticmethod
    def _rip(keyword_string):
        keyword_string = unidecode(keyword_string)
        keyword_string = keyword_string.replace(',', '')
        keyword_string = keyword_string.replace("'", '').replace('"', '')
        return keyword_string

    @property
    def affiliations(self):
        """
        The property, that will return the list of all affiliation ids, which the authors of the publication have.

        :return: A list of int affiliation ids
        """
        affiliation_list = []
        # Looping through the authors and adding all the affiliations, that have not already been added
        for author in self.authors:
            affiliation_list_difference = list(set(author.affiliations) - set(affiliation_list))
            affiliation_list += affiliation_list_difference

        return affiliation_list

    def get_id(self):
        """
        Returns the int scopus id of the publication.

        :return: int id
        """
        return int(self.id)

    def __int__(self):
        """
        Function to convert the object into an integer type. Returns the in scopus id, that identifies the publication.

        :return:The integer scopus id of the publication
        """
        return self.get_id()

    def contains_author(self, scopus_author):
        """
        Whether or not the given author has participated in this publication.

        :param scopus_author: The ScopusAuthor object representing the author, for which to check if he has worked on
            the publication
        :return: boolean
        """
        for author in self.authors:

            # Comparing the id of the author object given with the author objects in the list of authors for the
            # publication.
            if int(author) == int(scopus_author):
                return True

        # Returning False if True has not been returned in any of the previous iterations
        return False

    def contains_keyword(self, keyword):
        """
        Whether or not the publication is tagged with the given keyword.

        :param keyword: The string keyword for which to check the publication
        :return: boolean
        """
        return keyword in self.keywords

    def contains_affiliation(self, affiliation_id):
        """
        Whether ot not the publication is affiliated with the given affiliation id.

        Assembles a list of all different affiliation ids associated with all the authors of the publication and checks
        if the given affiliation id is in this list.
        :param affiliation_id: The int affiliation id to check for
        :return: boolean
        """
        return affiliation_id in self.affiliations

    def to_dict(self):

        author_dict_list = []
        for author in self.authors:
            author_dict = author.to_dict()
            author_dict_list.append(author_dict)

        dictionary = {
            'id': self.id,
            'eid': self.eid,
            'doi': self.doi,
            'title': self.title,
            'description': self.description,
            'date': self.date,
            'journal': self.journal,
            'volume': self.volume,
            'creator': self.creator.to_dict(),
            'keywords': self.keywords,
            'authors': author_dict_list,
            'citations': self.citations
        }
        return dictionary

    @staticmethod
    def from_dict(dictionary):
        if 'type' in dictionary and dictionary['type'] == 'ScopusPublication':
            author_list = []
            for author_dict in dictionary['authors']:
                author = ScopusAuthor.from_dict(author_dict)
                author_list.append(author)

            publication = ScopusPublication(
                dictionary['id'],
                dictionary['eid'],
                dictionary['doi'],
                dictionary['title'],
                dictionary['description'],
                dictionary['date'],
                ScopusAuthor.from_dict(dictionary['creator']),
                author_list,
                dictionary['citations'],
                dictionary['keywords'],
                dictionary['journal'],
                dictionary['volume']
            )
            return publication

        else:
            error_message = (
                'The type of the dict {} is not compatible for a ScopusAuthor conversion'
            ).format(
                str(dictionary)
            )
            raise TypeError(error_message)


class ScopusAuthor(ScopusIdentifierInterface, DictionaryConversionInterface):
    """
    The representation of the Author data saved with a publication, that was retrieved from the scopus database.

    This Author representation only contains very basic information such as the scopus author id, the first name, the
    last name and the list of affiliations. Objects of this kind will only be created from the data of a publication
    retrieval.
    The data for a author varies greatly in the scopus database and one cannot assume to get legitimate information
    from a author retrieval.
    """
    def __init__(self, first_name, last_name, id, affiliation_list):
        # Init the interface
        ScopusIdentifierInterface.__init__(self)
        DictionaryConversionInterface.__init__(self)

        self.id = id
        self.first_name = unidecode(first_name).replace("'", "")
        self.last_name = unidecode(last_name).replace("'", "")
        self.id = id
        self.affiliations = affiliation_list
        if '' in self.affiliations:
            self.affiliations.remove('')

    def get_id(self):
        """
        Returns the int author id of the author.

        :return: int
        """
        return int(self.id)

    def __int__(self):
        """
        Function to convert an object into an integer. Returns the int author id of the author.

        :return: int
        """
        return self.get_id()

    def __hash__(self):
        return hash('{} {}'.format(self.first_name, self.last_name))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def contains_affiliation(self, affiliation_id):
        """
        Whether the author is affiliated with the given affiliation id.

        :param affiliation_id: The int affiliation id to check for
        :return:
        """
        return int(affiliation_id) in self.affiliations

    def to_dict(self):
        """
        Turns the data inside the object into a dict

        :return: a dict object containing all the data of the object
        """
        dictionary = {
            'type': 'ScopusAuthor',
            'id': self.id,
            'first_name': self.first_name,
            'last_name': self.last_name,
            'affiliations': self.affiliations
        }
        return dictionary

    @staticmethod
    def from_dict(dictionary):
        """
        Loads a ScopusAuthor object from a dictionary, that has the correct format.

        :param dictionary: The dict to convert into a ScopusAuthor object
        :return: A scopusAuthor object
        """
        if 'type' in dictionary.keys() and dictionary['type'] == 'ScopusAuthor':
            author = ScopusAuthor(
                dictionary['first_name'],
                dictionary['last_name'],
                dictionary['id'],
                dictionary['affiliations']
            )
            return author
        else:
            error_message = (
                'The type of the dict {} is not compatible for a ScopusAuthor conversion'
            ).format(
                str(dictionary)
            )
            raise TypeError(error_message)


class ScopusAuthorProfile(ScopusIdentifierInterface):
    """
    Represents the basic data structure from a scopus author retrieval.
    """
    def __init__(self, author_id, first_name, last_name, h_index, citation_count, document_count, publication_list):
        # Init the reference
        ScopusIdentifierInterface.__init__(self)

        self.id = author_id
        self.first_name = unidecode(first_name).replace("'", '')
        self.last_name = unidecode(last_name).replace("'", '')
        self.h_index = h_index
        if h_index is None:
            self.h_index = 0
        self.citation_count = citation_count
        if citation_count is None:
            self.citation_count = 0
        self.document_count = document_count
        if document_count is None:
            self.document_count = 0
        self.publications = publication_list
        if '' in self.publications:
            self.publications.remove('')

    def get_id(self):
        """
        Returns the int author id of the author.

        :return: int
        """
        return int(self.id)

    def __int__(self):
        """
        Method to convert an object to int. Returns the int author id of the author.

        :return: int
        """
        return self.get_id()

    def contains_publication(self, scopus_publication):
        """
        whether or not the author has contributed to the given publication.

        :param scopus_publication: The ScopusPublication object to check for contribution of the author
        :return: boolean
        """
        for scopus_id in self.publications:
            if scopus_id == int(scopus_publication):
                return True

        return False


class ScopusAffiliation(ScopusIdentifierInterface):
    """
    Represents the basic information about a scopus affiliation retrieval.
    """
    def __init__(self, affiliation_id, country, city, institute):
        # Init the Interface
        ScopusIdentifierInterface.__init__(self)

        self.id = affiliation_id
        self.country = country
        self.city = city
        self.institute = institute

    def get_id(self):
        """
        The int affiliation id.

        :return: int
        """
        return int(self.id)

    def __int__(self):
        """
        The method to convert the object to int. Returns the int affiliation id of the affiliation.

        :return: int
        """
        return self.get_id()


class ScopusAuthorObservation:

    def __init__(self, author_id_list, first_name, last_name, keyword_list, whitelist, blacklist):

        self.ids = author_id_list
        self.first_name = first_name
        self.last_name = last_name
        self.keywords = keyword_list
        self.whitelist = whitelist
        self.blacklist = blacklist

    def whitelist_contains(self, affiliation_id):
        """
        Whether or not the affiliation id is in the affiliation whitelist for that author.

        :param affiliation_id: The int id to check for
        :return: boolean
        """
        try:
            return int(affiliation_id) in self.whitelist
        except ValueError:
            return False

    def whitelist_contains_any(self, affiliation_id_list):
        """
        Whether or not any one of the affiliation ids of the given list is in the whitelist for that author.

        :param affiliation_id_list: A list of int affiliation ids to check
        :return: boolean
        """
        for affiliation_id in affiliation_id_list:
            if self.whitelist_contains(affiliation_id):
                return True
        return False

    def whitelist_check_publication(self, scopus_publication):
        """
        Whether the observed author is author of the given publication and if one of the affiliations of the
        publication is in the affiliation whitelist.

        :param scopus_publication: The ScopusPublication to check
        :return: boolean
        """
        # Checks whether the observed author is actually an author of the given pub and then if any of the affiliations
        # of the author is in the whitelist
        _contains = False
        for id in self.ids:
            if scopus_publication.contains_author(id):
                _contains = True
                break
        return _contains and self.whitelist_contains_any(scopus_publication.affiliations)

    def blacklist_contains(self, affiliation_id):
        """
        Whether the given affiliation id in the affiliation blacklist.

        :param affiliation_id: The int id
        :return: boolean
        """
        try:
            return int(affiliation_id) in self.blacklist
        except ValueError:
            return False

    def blacklist_contains_any(self, affiliation_id_list):
        """
        Whether one of the affiliation ids of the given list is in the blacklist for that author.

        :param affiliation_id_list: A list of affiliation ids to check
        :return: boolean
        """
        for affiliation_id in affiliation_id_list:
            if self.blacklist_contains(affiliation_id):
                return True
        return False

    def blacklist_check_publication(self, scopus_publication):
        """
        Whether the observed author is author of the given publication and if one of the affiliation of the
        publication is in the affiliation blacklist.

        :param scopus_publication: The ScopusPublication to check
        :return: boolean
        """
        # Checks whether the observed author is actually an author of the given pub and then if any of the affiliations
        # of the author is in the blacklist
        _contains = False
        for id in self.ids:
            if scopus_publication.contains_author(id):
                _contains = True
                break
        return _contains and self.blacklist_contains_any(scopus_publication.affiliations)

